# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Class for managing the Excel Application Connection and common read/write operations."""

from typing import Optional, List, Set, Callable, Any, Union, Dict
from contextlib import contextmanager
import os
import pathlib

from PHX.xl import xl_data
from PHX.xl.xl_typing import (
    xl_Framework_Protocol,
    xl_Book_Protocol,
    xl_Books_Protocol,
    xl_Sheets_Protocol,
    xl_Sheet_Protocol,
    xl_Range_Protocol,
    xl_app_Protocol,
    xl_apps_Protocol,
)

# -----------------------------------------------------------------------------
# -- Exceptions


class ReadRowsError(Exception):
    def __init__(self, row_start, row_end):
        self.msg = (
            f"Error: row_start should be less than "
            f"row_end. Got {row_start}, {row_end}"
        )
        super().__init__(self.msg)


class NoActiveExcelRunningError(Exception):
    def __init__(self):
        self.msg = (
            "\n\tError: No active instance of Excel running?"
            "\n\tPlease open Excel and try again."
        )
        super().__init__(self.msg)


class ReadMultipleColumnsError(Exception):
    def __init__(self, _c1, _c2):
        self.msg = (
            f'\n\tError: Cannot use "read_multiple_columns()" with _col_start={_c1}'
            f'and _col_end={_c2}. Please use "read_single_column()" instead.'
        )
        super().__init__(self.msg)


class WriteValueError(Exception):
    def __init__(self, _value, _range, _worksheet, _e):
        self.msg = (
            f"\n\n\tSomething went wrong trying to write the value: '{_value}' to the "
            f"cell: '{_range}' on worksheet: '{_worksheet}'. Please make sure that a "
            f"valid Excel file is open, and both the worksheet and workbook are "
            f"unprotected.\n\n{_e}"
        )
        super().__init__(self.msg)


class XlReadException(Exception):
    def __init__(self, _range: str):
        self.msg = (
            f"\n\tError: 'get_single_data_item()' can only be used on a"
            f"single range. Got range: '{_range}'. Please use 'get_data()' instead."
        )
        super().__init__(self.msg)


class NoSuchFileError(Exception):
    def __init__(self, _file: pathlib.Path):
        self.msg = f"\n\tError: Cannot locate the file \n{_file}?"
        super().__init__(self.msg)


# -----------------------------------------------------------------------------


def silent_print(_input: Any) -> None:
    """Default 'output' for XLConnection."""
    return


# -----------------------------------------------------------------------------


class XLConnection:
    """An Excel connection Facade / Interface."""

    def __init__(
        self,
        xl_framework,
        output: Callable[[str], Any] = silent_print,
        xl_file_path: Optional[pathlib.Path] = None,
    ) -> None:
        """Facade class for Excel Interop

        Arguments:
        ----------
            * xl (xl_Framework_Protocol): The Excel framework to use to interact with XL.

            * output (Callable[[str], Any]): The output function to use. Default is silent
                (no output), or provide the standard python 'print' for standard-out, etc.

            * xl_file_path (Optional[pathlib.Path]): The full path to an XL file to use. If none,
                will default to the 'active' xl workbook.
        """
        # -- Note: can not type-hint xl_framework in the Class argument line
        # -- cus' Python-3.7 doesn't have Protocols yet. It does see to work
        # -- when type-hinting the actual attribute though.
        self.xl: xl_Framework_Protocol = xl_framework
        self._output: Callable[[str], Any] = output
        self.xl_file_path: Optional[pathlib.Path] = xl_file_path

        self._wb: Optional[xl_Book_Protocol] = None
        self.output(f"> connected to excel doc: '{self.wb.fullname}'")

    def activate_new_workbook(self) -> xl_Book_Protocol:
        """Create a new blank workbook and set as the 'Active' book. Returns the new book."""
        new_book = self.books.add()
        self._wb = new_book
        return new_book

    @property
    def excel_running(self) -> bool:
        """Returns True if Excel is currently running, False if not"""
        return self.apps.count > 0

    def start_excel_app(self) -> Optional[xl_app_Protocol]:
        """Starts Excel Application, if it is not currently running."""
        if self.excel_running:
            return None

        print("adding a new app to the self.apps....")
        new_ap = self.apps.add()
        new_ap.visible = True

        if os.name == "nt":
            # Need to add a default workbook on Windows
            new_ap.books.add()

        return new_ap

    @property
    def apps(self) -> xl_apps_Protocol:
        """Return the xl framework 'apps' object."""
        return self.xl.apps

    @property
    def books(self) -> xl_Books_Protocol:
        """Return the xl framework 'books' object."""
        return self.xl.books

    @property
    def os_is_windows(self) -> bool:
        """Return True if the current OS is Windows. False if it is Mac/Linux"""
        return os.name == "nt"

    def get_workbook(self) -> xl_Book_Protocol:
        """Return the right Workbook, depending on the App state and user inputs."""
        if not self.excel_running:
            print("no Excel running, staring a new Application instance.")
            self.start_excel_app()

        # -- If a specific file path is provided, open that one
        if self.xl_file_path:
            if not self.xl_file_path.exists():
                raise NoSuchFileError(self.xl_file_path)
            return self.books.open(self.xl_file_path)

        # -- If no books are open yet, create a new one and return it
        if self.books.count == 0:
            return self.books.add()

        # -- Otherwise, just return the Active workbook
        return self.books.active

    @property
    def wb(self) -> xl_Book_Protocol:
        """Returns the Workbook of the active Excel Instance."""
        if self._wb is not None:  # -- Cache
            return self._wb
        else:
            self._wb = self.get_workbook()
            return self._wb

    def autofit_columns(self, _sheet_name: str) -> None:
        """Runs autofit on all the columns in a sheet."""
        sht = self.get_sheet_by_name(_sheet_name)
        sht.autofit(axis="c")  # by-columns

    def autofit_rows(self, _sheet_name: str) -> None:
        """Runs autofit on the rows in a sheet."""
        sht = self.get_sheet_by_name(_sheet_name)
        # sht.activate()
        sht.autofit(axis="r")  # by-rows

    def clear_range_data(self, _sheet_name: str, _range: str) -> None:
        """Sets the specified excel sheet's range to 'None'

        Arguments:
        ---------
            * _sheet_name: (str) The name of the worksheet
            * _range: (str) The cell range to write to (ie: "A1") or a set of ranges (ie: "A1:B4")
        """
        self.get_sheet_by_name(_sheet_name).range(_range).value = None

    def clear_sheet_contents(self, _sheet_name: str) -> None:
        """Clears the content of the whole sheet but leaves the formatting."""
        self.get_sheet_by_name(_sheet_name).clear_contents()

    def clear_sheet_formats(self, _sheet_name: str) -> None:
        """Clears the format of the whole sheet but leaves the content."""
        self.get_sheet_by_name(_sheet_name).clear_formats()

    def clear_sheet_all(self, _sheet_name: str) -> None:
        """Clears the content and formatting of the whole sheet."""
        self.get_sheet_by_name(_sheet_name).clear()

    def create_new_worksheet(
        self, _sheet_name: str, before: Optional[str] = None, after: Optional[str] = None
    ) -> None:
        """Try and add a new Worksheet to the Workbook."""
        try:
            self.wb.sheets.add(_sheet_name, before, after)
            self.output(f"Adding '{_sheet_name}' to Workbook")
        except ValueError:
            self.output(f"Worksheet '{_sheet_name}' already in Workbook.")

        self.get_sheet_by_name(_sheet_name).clear()

    def find_row(
        self,
        _search_item: xl_data.xl_range_single_value,
        _data: List[xl_data.xl_range_single_value],
        _start: int = 0,
    ) -> Optional[int]:
        for i, _ in enumerate(_data, start=_start):
            if _search_item == _:
                return i
        return None

    def get_row_num_of_value_in_column(
        self, sheet_name: str, row_start: int, row_end: int, col: str, find: str
    ) -> Optional[int]:
        """Returns the row number of the first instance of a specific value
        found within a column, or None if not found.

        Arguments:
        ----------
            * sheet_name: (str) The name of the sheet to be looking in
            * row_start: (int) The row number to begin looking from
            * row_end: (int) The row number to stop looking on
            * col: (str) The column to look in
            * find: (Optional[str]) The string to search for (or 'None' for blank cell)

        Returns:
        --------
            * Optional[int]: The row number where the first instance of the specified
                value is found or None if no instances are found.
        """

        if row_start > row_end:
            raise ReadRowsError(row_start, row_end)

        row: int = row_start
        sh: xl_Sheet_Protocol = self.get_sheet_by_name(sheet_name)

        while row <= row_end:
            if sh.range(f"{col}{row}").value != find:
                row += 1
            else:
                return row
        return None

    def get_worksheet_names(self) -> Set[str]:
        """Return a set of all the worksheet names in the workbook.

        Returns:
        --------
            * (Set[str])
        """
        return {sh.name.upper() for sh in self.wb.sheets}

    def get_sheet_by_name(self, _sheet_name: Union[str, int]) -> xl_Sheet_Protocol:
        """Returns an Excel Sheet with the specified name, or KeyError if not found.

        Arguments:
        ----------
            * _sheet_name: (Union[str, int]): The excel sheet name or index num. to locate.

        Returns:
        --------
            * (xw.main.Sheet): The excel sheet found.
        """
        if str(_sheet_name).upper() not in self.get_worksheet_names():
            msg = f"Error: Key '{_sheet_name}' was not found in the Workbook '{self.wb.name}' Sheets?"
            raise KeyError(msg)

        return self.wb.sheets[_sheet_name]

    def get_last_sheet(self) -> xl_Sheet_Protocol:
        """Return the last Worksheet in the Workbook.

        Returns:
        --------
            * (xl_Sheet_Protocol) The last Worksheet in the Workbook.
        """
        return self.wb.sheets[-1]

    def get_last_used_row_num_in_column(self, _sheet_name: str, _col: str) -> int:
        """Return the row number of the last cell in a column with a value in it.

        Arguments:
        ----------
            * _sheet_name (str): The name of the Worksheet to read from.
            * _col (str): The Alpha character of the column to read.

        Returns:
        --------
            * (int): The number of the last row in the column with a value.
        """
        sheet = self.get_sheet_by_name(_sheet_name)
        # sheet.activate()
        col_range = sheet.range(f"{_col}:{_col}")
        col_last_cell_range = sheet.range(col_range.last_cell.address)
        group_last_cell_range = col_last_cell_range.end("up")  # same as 'Ctrl-Up'
        return group_last_cell_range.row

    def get_single_column_data(
        self,
        _sheet_name: str,
        _col: str,
        _row_start: Optional[int] = None,
        _row_end: Optional[int] = None,
    ) -> List[xl_data.xl_range_single_value]:
        """Return a list with the values read from a single column of the excel document.

        Arguments:
        ----------
            * _sheet_name: (str) The Excel Worksheet to read from.
            * _col: (str) The Column letter to read.
            * _row_start: (Optional[int]) default=None
            * _row_end: (Optional[int]) default=None
        Returns:
        --------
            (List[xl_range_value]): The data from Excel worksheet, as a list.
        """

        if not _row_start:
            _row_start = 1
        if not _row_end:
            _row_end = self.get_last_used_row_num_in_column(_sheet_name, _col)

        address = f"{_col}{_row_start}:{_col}{_row_end}"
        self.output(f"Reading: '{address}' data on sheet: '{_sheet_name}'")

        sheet = self.get_sheet_by_name(_sheet_name)
        # sheet.activate()
        col_range = sheet.range(f"{address}")
        return col_range.value  # type: ignore

    def get_last_used_column_in_row(self, _sheet_name: str, _row: int) -> str:
        """Return the column letter of the last cell in a column with a value in it.

        Arguments:
        ----------
            * _sheet_name (str): The name of the Worksheet to read from.
            * _col (str): The Alpha character of the column to read.

        Returns:
        --------
            * (str): The Letter of the last column in the row with a value.
        """
        sheet = self.get_sheet_by_name(_sheet_name)
        # sheet.activate()
        row_range = sheet.range(f"{_row}:{_row}")
        row_last_cell_range = sheet.range(row_range.last_cell.address)
        group_last_cell_range = row_last_cell_range.end("left")  # same as 'Ctrl-Left'
        return xl_data.xl_chr(xl_data.xl_ord("A") + group_last_cell_range.column - 1)

    def get_single_row_data(
        self, _sheet_name: str, _row_number: int
    ) -> List[xl_data.xl_range_single_value]:
        """Return all the data from a single Row in the Excel Workbook.

        Arguments:
        ----------
            * _sheet_name (str): The name of the sheet to read
            * _row_number (int): The row number to read

        Returns:
        --------
            * (List[xl_data.xl_range_single_value]) A List of the data read from XL.
        """

        self.output(f"Reading: Row-{_row_number} on sheet: '{_sheet_name}'")

        sheet = self.get_sheet_by_name(_sheet_name)
        # sheet.activate()
        last_col_letter = self.get_last_used_column_in_row(_sheet_name, _row_number)
        row_range = sheet.range(f"A{_row_number}:{last_col_letter}{_row_number}")
        return row_range.value  # type: ignore

    def get_multiple_column_data(
        self,
        _sheet_name: str,
        _col_start: str,
        _col_end: str,
        _row_start: int = 1,
        _row_end: int = 100,
    ) -> xl_data.xl_range_list2D_value:
        """Return a list of lists with the values read from a specified block of the xl document.

        Arguments:
        ----------
            * _sheet_name: (str) The Worksheet to read from.
            * _col_start: (str) The Column letter to read from.
            * _col_end: (str) The Column letter to read to.
            * _row_start: (int) default=1
            * _row_end: (int) default=100

        Returns:
        --------
            (List[List[xl_range_value]]): The data from Excel worksheet, as a list of lists.
        """

        if _col_start == _col_end:
            raise ReadMultipleColumnsError(_col_start, _col_end)

        # -- Use xl.Range() instead of ord() since ord('KL') will fail
        rng: xl_Range_Protocol = self.xl.Range(f"{_col_end}1:{_col_start}1")
        _ndim = len(rng.columns)

        sheet = self.get_sheet_by_name(_sheet_name)
        rng = sheet.range(f"{_col_start}{_row_start}:{_col_end}{_row_end}")
        rng = rng.options(ndim=_ndim)
        return rng.value

    def get_data(self, _sheet_name: str, _range: str) -> xl_data.xl_writable:
        """Return a value or values from the Excel document.

        Arguments:
        ----------
            * _sheet_name: (str) The name of the worksheet to read from.
            * _range: (str) The cell range to read from (ie: "A1") or a set of ranges (ie: "A1:B4")

        Returns:
        ---------
            * (xl_writable): The resultant value/values returned from excel.
        """
        self.output(f"Reading: {_sheet_name}:{_range}")
        return self.get_sheet_by_name(_sheet_name).range(_range).value

    def get_single_data_item(
        self, _sheet_name: str, _range: str
    ) -> xl_data.xl_range_single_value:
        """Return a single value from the Excel document.

        Arguments:
        ----------
            * _sheet_name: (str) The name of the worksheet to read from.
            * _range: (str) The cell range to read from to (ie: "A1")

        Returns:
        ---------
            * (xl_writable): The resultant value returned from excel.
        """

        if ":" in _range:
            raise XlReadException(_range)
        self.output(f"Reading: {_sheet_name}:{_range}")
        return self.get_sheet_by_name(_sheet_name).range(_range).value  # type: ignore

    def get_data_by_columns(
        self, _sheet_name: str, _range_address: str
    ) -> List[List[xl_data.xl_range_single_value]]:
        """Returns a List of column data, each column in a list.
        ie: "A1:D12" -> [[A1, A2, ... A12], [B1, B2, ... B12], ... [D1, D2, ... D12]]

        Arguments:
        ----------
            * _sheet_name: (str) The name of the worksheet to read from
            * _range_address: (str) The range read. ie: "A1:D56"

        Returns:
        --------
            * (List[List[xl_data.xl_range_single_value]])
        """
        sheet = self.get_sheet_by_name(_sheet_name)
        rng = sheet.range(_range_address).options(transpose=True)
        return rng.value  # type: ignore

    def get_data_with_column_letters(
        self, _sheet_name: str, _range_address: str
    ) -> Dict[str, List[xl_data.xl_range_single_value]]:
        """Returns a Dict of column data, key'd by the Column Letter.
        ie: "A1:D12"->{"A":[A1, A2,...A12], "B":[...], ..."D":[D1., D2,...D12]}

        Arguments:
        ----------
            * _sheet_name: (str) The name of the worksheet to read from
            * _range_address: (str) The range read. ie: "A1:D56"

        Returns:
        --------
            * (Dict[str, List[xl_data.xl_range_single_value]])
        """

        raw_data = self.get_data_by_columns(_sheet_name, _range_address)

        # -- Shape the data into a dict, key'd by Column Letter
        # -- ie: {"P":["some", ...], "Q":["data", ...], ...}
        rng = self.xl.Range(_range_address)
        start_col_letter = xl_data.xl_col_num_as_chr(rng.column)
        start_col_number = xl_data.xl_ord(start_col_letter)
        return {
            xl_data.xl_chr(i): data_list
            for i, data_list in enumerate(raw_data, start=start_col_number)
        }

    def group_rows(self, _sheet_name: str, _row_start: int, _row_end: int) -> None:
        """Group one or more rows."""
        sht = self.get_sheet_by_name(_sheet_name)
        # sht.activate()
        if self.os_is_windows:
            sht.api.Rows(f"{_row_start}:{_row_end}").Group()
        else:
            sht.api.rows[f"{_row_start}:{_row_end}"].group()
        return None

    def hide_group_details(self, _sheet_name: str) -> None:
        """Hide (collapse) all the 'Groups' on the specified worksheet."""
        sheet = self.get_sheet_by_name(_sheet_name)
        if self.os_is_windows:
            pass
        else:
            sheet.api.outline_object.show_levels(row_levels=1)

    @contextmanager
    def in_silent_mode(self):
        """Context Manager which turns off screen-refresh and auto-calc in the
        Excel App in order to help speed up read/write. Turns back on screen-refresh
        and auto-calc in the Excel App when done or on any error.
        """
        try:
            self.wb.app.screen_updating = False
            self.wb.app.display_alerts = False
            self.wb.app.calculation = "manual"
            yield
        finally:
            self.wb.app.screen_updating = True
            self.wb.app.display_alerts = True
            self.wb.app.calculation = "automatic"
            self.wb.app.calculate()

    def output(self, _input):
        """Used to set the output method. Default is None (silent).

        Arguments:
        ----------
            * _input (SupportsString): The string to output.

        Returns:
        --------
            * (None)
        """
        try:
            self._output(str(_input))
        except:
            # If _input=None, ignore...
            pass

    def unprotect_all_sheets(self) -> None:
        """Walk through all the sheets and unprotect them all."""
        for sheet in self.wb.sheets:
            if self.os_is_windows:
                sheet.api.Unprotect()
            else:
                sheet.api.unprotect()

    def write_xl_item(
        self,
        _xl_item: Union[xl_data.XlItem, xl_data.XLItem_List],
        _transpose: bool = False,
    ) -> None:
        """Writes a single XLItem to the worksheet

        Arguments:
        ---------
            * _xl_item: (XLItem) The XLItem with a sheet_name, range and value to write.
            * _transpose: (bool) Transpose the data before writing. Default=False. Set
                to true if you are passing in a list of items and want them to be
                written as rows instead of as columns.
        """

        self.output(
            f"Writing: {_xl_item.sheet_name}:{_xl_item.xl_range}={_xl_item.write_value}"
        )

        try:
            xl_sheet = self.get_sheet_by_name(_xl_item.sheet_name)
            xl_range = xl_sheet.range(_xl_item.xl_range).options(transpose=_transpose)
            xl_range.value = _xl_item.write_value  # type: ignore

            if _xl_item.has_color:
                if _xl_item.value_is_iterable:
                    # -- color all the columns
                    for i in range(len(_xl_item.write_value)):  # type: ignore
                        _xl_range = xl_range.offset(column_offset=i)
                        _xl_range.color = _xl_item.range_color
                        _xl_range.font.color = _xl_item.font_color
                else:
                    xl_range.color = _xl_item.range_color
                    xl_range.font.color = _xl_item.font_color

        except AttributeError as e:
            raise AttributeError(e)
        except Exception as e:
            raise WriteValueError(
                _xl_item.write_value, _xl_item.xl_range, _xl_item.sheet_name, e
            )



# -*- coding: utf-8 -*-
# -*- Python Version: 2.7 -*-

import pytest
from ph_units.converter import convert, unit_type_dict


def test_W():
    assert convert(1, "W", "KW") == 0.001
    assert convert(1, "W", "W") == 1
    assert convert(1, "W", "BTU/HR") == 3.412141156
    assert convert(1, "W", "KBTU/HR") == 0.003412141


def test_kW():
    assert convert(1, "KW", "KW") == 1
    assert convert(1, "KW", "W") == 1_000
    assert convert(1, "KW", "BTU/HR") == 3412.141156
    assert convert(1, "KW", "KBTU/HR") == 3.412141156


def test_Btu_hr():
    assert convert(1, "BTUH", "W") == 0.293071111
    assert convert(1, "BTUH", "KW") == 0.000293071
    assert convert(650_678, "BTUH", "KW") == 190.694852138


def test_kBtu_aliases():
    assert convert(1, "KBTU/HR", "KW") == 0.293071111
    assert convert(1, "KBTU/H", "KW") == 0.293071111
    assert convert(1, "KBTUHR", "KW") == 0.293071111


def test_kBtu_hr():
    assert convert(1, "KBTUH", "BTUH") == 1_000
    assert convert(1, "KBTUH", "W") == 293.0711111
    assert convert(1, "KBTUH", "KW") == 0.293071111



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""PHX Foundation Class"""

from typing import Union, Type, Optional
from dataclasses import dataclass
from PHX.model.enums.foundations import (
    FoundationType,
    CalculationSetting,
    PerimeterInsulationPosition,
)


@dataclass
class PhxFoundation:
    display_name: str = "__unnamed_foundation__"

    foundation_setting_num: CalculationSetting = CalculationSetting.USER_DEFINED
    _foundation_type_num: FoundationType = FoundationType.NONE

    @property
    def foundation_type_num(self) -> FoundationType:
        return self._foundation_type_num

    @foundation_type_num.setter
    def foundation_type_num(self, _input: Union[FoundationType, int]) -> None:
        if isinstance(_input, FoundationType):
            self._foundation_type_num = _input
        self._foundation_type_num = FoundationType(_input)


class PhxHeatedBasement(PhxFoundation):
    def __init__(self):
        super().__init__()
        self.floor_slab_area_m2: Optional[float] = 0.0
        self.floor_slab_u_value: Optional[float] = 1.0
        self.floor_slab_exposed_perimeter_m: Optional[float] = 0.0
        self.slab_depth_below_grade_m: Optional[float] = 2.5
        self.basement_wall_u_value: Optional[float] = 1.0


class PhxUnHeatedBasement(PhxFoundation):
    def __init__(self):
        super().__init__()
        self.floor_ceiling_area_m2: Optional[float] = 0.0
        self.ceiling_u_value: Optional[float] = 0.0
        self.floor_slab_exposed_perimeter_m: Optional[float] = 0.0
        self.slab_depth_below_grade_m: Optional[float] = 0.0
        self.basement_wall_height_above_grade_m: Optional[float] = 0.0
        self.basement_wall_uValue_below_grade: Optional[float] = 0.0
        self.basement_wall_uValue_above_grade: Optional[float] = 0.0
        self.floor_slab_u_value: Optional[float] = 0.0
        self.basement_volume_m3: Optional[float] = 0.0
        self.basement_ventilation_ach: Optional[float] = 0.0


class PhxSlabOnGrade(PhxFoundation):
    def __init__(self):
        super().__init__()
        self.floor_slab_area_m2: Optional[float] = 0.0
        self.floor_slab_u_value: Optional[float] = 1.0
        self.floor_slab_exposed_perimeter_m: Optional[float] = 0.0
        self._perim_insulation_position: PerimeterInsulationPosition = (
            PerimeterInsulationPosition.VERTICAL
        )
        self.perim_insulation_width_or_depth_m: Optional[float] = 0.300
        self.perim_insulation_thickness_m: Optional[float] = 0.050
        self.perim_insulation_conductivity: Optional[float] = 0.04

    @property
    def perim_insulation_position(self) -> PerimeterInsulationPosition:
        return self._perim_insulation_position

    @perim_insulation_position.setter
    def perim_insulation_position(
        self, _input: Union[PerimeterInsulationPosition, int, None]
    ) -> None:
        if not _input:
            return
        if isinstance(_input, PerimeterInsulationPosition):
            self._perim_insulation_position = _input
        self._perim_insulation_position = PerimeterInsulationPosition(_input)


class PhxVentedCrawlspace(PhxFoundation):
    def __init__(self):
        super().__init__()
        self.crawlspace_floor_slab_area_m2: Optional[float] = 0.0
        self.ceiling_above_crawlspace_u_value: Optional[float] = 1.0
        self.crawlspace_floor_exposed_perimeter_m: Optional[float] = 0.0
        self.crawlspace_wall_height_above_grade_m: Optional[float] = 0.0
        self.crawlspace_floor_u_value: Optional[float] = 1.0
        self.crawlspace_vent_opening_are_m2: Optional[float] = 0.0
        self.crawlspace_wall_u_value: Optional[float] = 1.0


# type alias
PhxFoundationTypes = Union[
    Type[PhxFoundation],
    Type[PhxHeatedBasement],
    Type[PhxUnHeatedBasement],
    Type[PhxSlabOnGrade],
    Type[PhxVentedCrawlspace],
]



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""PHX Lighting Load."""

from __future__ import annotations
from typing import ClassVar, Any
from dataclasses import dataclass, field


@dataclass
class PhxLoadLighting:
    """A PHX Load for the Lighting."""

    installed_w_per_m2: float = 0.0

    @property
    def unique_key(self):
        return "{}_{:.4f}_".format(self.__class__.__name__, self.installed_w_per_m2)



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""DEV SANDBOX: convert a WUFI XML file over to HBJSON."""

import pathlib

from rich import print
from PHX.from_WUFI_XML.read_WUFI_XML_file import get_WUFI_XML_file_as_dict
from PHX.from_WUFI_XML.phx_converter import convert_WUFI_XML_to_PHX_project
from PHX.to_WUFI_XML import xml_builder, xml_txt_to_file

from PHX.from_WUFI_XML.wufi_file_schema import WUFIplusProject

SOURCE_DIR = pathlib.Path("tests", "_reference_xml")
SOURCE_FILE_NAMES = [
    "test_apartment.xml",
]
TARGET_DIR = pathlib.Path("tests", "_regenerated_xml")

print(" -" * 50)
for i, xm_source_file_name in enumerate(SOURCE_FILE_NAMES):
    print(f" -" * 25)
    xm_source_file = SOURCE_DIR / xm_source_file_name

    # ----------------------------------------------------------------
    # -- Read in the WUFI-XML File as a new Pydantic Model
    print(f"[green bold]> Reading in the file: {xm_source_file}[/green bold]")
    wufi_xml_data = get_WUFI_XML_file_as_dict(xm_source_file)
    wufi_xml_model = WUFIplusProject.parse_obj(wufi_xml_data)

    # with open(f"dict_{i}.txt", "w") as f:
    #     print(wufi_xml_data, file=f)

    # with open(f"model_{i}.txt", "w") as f:
    #     print(wufi_xml_model, file=f)

    # ----------------------------------------------------------------
    # Convert the Pydnatic WUFI model over to a PHX model
    print(f"[green bold]> Converting XML file to PHX Model[/green bold]")
    phx_project = convert_WUFI_XML_to_PHX_project(wufi_xml_model)

    # # # # --- Output the PHX model to a WUFI-XML
    target_file = TARGET_DIR / xm_source_file_name
    xml_txt = xml_builder.generate_WUFI_XML_from_object(phx_project)

    print(f"[bold]> Saving the XML file to: ./{target_file}[/bold]")
    xml_txt_to_file.write_XML_text_file(target_file, xml_txt, False)



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Functions used to create Project elements from the Honeybee-Model"""

from typing import Optional

from honeybee import model, room

from honeybee_energy.schedule import ruleset as hbe_ruleset
from honeybee_energy.properties.room import RoomEnergyProperties
from honeybee_energy.lib.scheduletypelimits import schedule_type_limit_by_identifier

from honeybee_energy_ph.properties import ruleset as phx_ruleset
from honeybee_ph_utils.schedules import calc_four_part_vent_sched_values_from_hb_room

from PHX.model import project
from PHX.model.schedules import ventilation, occupancy, lighting


def _room_has_ph_style_ventilation(_hb_room: room.Room) -> bool:
    """Returns True if the HB Room has detailed PH-Style ventilation schedule
        information, False if not.

    Arguments:
    ----------
        * _hb_room (room.Room): The Honeybee Room to look at.

    Returns:
    --------
        * (bool):
    """

    # -------------------------------------------------------------------------
    # -- Honeybee-Energy data might not be there...
    hbe_prop: Optional[RoomEnergyProperties] = _hb_room.properties.energy  # type: ignore
    if not hbe_prop:
        # No Honeybee-Energy Room Properties
        return False

    if hbe_prop.ventilation.schedule is None:
        # Not Honeybee-Energy Ventilation Schedule
        return False

    # -------------------------------------------------------------------------
    # -- Check Honeybee-Energy-PH detailed data
    hbe_vent_sched_prop = hbe_prop.ventilation.schedule.properties
    hbph_sched_prop: phx_ruleset.ScheduleRulesetPhProperties = hbe_vent_sched_prop.ph  # type: ignore
    if not hbph_sched_prop.daily_operating_periods:
        # No Honeybee-Energy-PH Schedule detailed Operating Periods
        return False

    # -- So then the room must have the detailed data
    return True


def _create_vent_schedule_from_hb_style(
    _hb_room: room.Room,
) -> ventilation.PhxScheduleVentilation:
    """Returns a new PHX Ventilation Schedule based on the HB-Room's E+
    HB-Style info found on the energy.ventilation

    This is used when no detailed PH-Style info is set by the user and you want to
    convert over an existing E+/HB style ventilation object to PH-Style.

    Arguments:
    ----------
        * _hb_room (room.Room): The Honeybee Room to build the new PHX UtilizationPatternVent from.

    Returns:
    --------
        * (ventilation.PhxScheduleVentilation): A new PHX Ventilation Schedule.
    """

    # -- Type Aliases
    hbe_room_prop: RoomEnergyProperties = _hb_room.properties.energy  # type: ignore

    new_phx_vent_schedule = ventilation.PhxScheduleVentilation()

    wufi_sched = calc_four_part_vent_sched_values_from_hb_room(_hb_room)
    op_periods = new_phx_vent_schedule.operating_periods
    op_periods.high.period_operating_hours = wufi_sched.high.period_operating_hours
    op_periods.high.period_operation_speed = wufi_sched.high.period_speed
    op_periods.standard.period_operating_hours = (
        wufi_sched.standard.period_operating_hours
    )
    op_periods.standard.period_operation_speed = wufi_sched.standard.period_speed
    op_periods.basic.period_operating_hours = wufi_sched.basic.period_operating_hours
    op_periods.basic.period_operation_speed = wufi_sched.basic.period_speed
    op_periods.minimum.period_operating_hours = wufi_sched.minimum.period_operating_hours
    op_periods.minimum.period_operation_speed = wufi_sched.minimum.period_speed

    # -- Keep all the IDs in alignment....
    new_phx_vent_schedule.identifier = hbe_room_prop.ventilation.schedule.identifier
    new_phx_vent_schedule.id_num = new_phx_vent_schedule._count
    ph_sched_props: phx_ruleset.ScheduleRulesetPhProperties = hbe_room_prop.ventilation.schedule.properties.ph  # type: ignore
    ph_sched_props.id_num = new_phx_vent_schedule.id_num  # <--- Important!
    new_phx_vent_schedule.name = hbe_room_prop.ventilation.schedule.display_name

    return new_phx_vent_schedule


def _create_vent_schedule_from_ph_style(
    _hb_room: room.Room,
) -> ventilation.PhxScheduleVentilation:
    """Returns a new PHX Utilization Pattern (Vent) based on the HB-Room's detailed
    PH-Style info found on the energy.ventilation.schedule.properties.ph

    Arguments:
    ----------
        * _hb_room (room.Room): The Honeybee Room to build the new PHX UtilizationPatternVent from.

    Returns:
    --------
        * (ventilation.PhxScheduleVentilation): A new PHX Ventilation Schedule
    """

    # -- Type Aliases
    hbe_room_prop: RoomEnergyProperties = _hb_room.properties.energy  # type: ignore
    hbe_vent_sched = hbe_room_prop.ventilation.schedule
    hbe_vent_sched_prop = hbe_vent_sched.properties
    hbe_vent_sched_prop_ph: phx_ruleset.ScheduleRulesetPhProperties = hbe_vent_sched_prop.ph  # type: ignore

    # -- Create the new Schedule object
    new_phx_vent_schedule = ventilation.PhxScheduleVentilation()
    new_phx_vent_schedule.name = hbe_room_prop.ventilation.schedule.display_name

    # -- Set all the ventilation schedule data from the room's properties
    new_phx_vent_schedule.operating_hours = 24.0
    new_phx_vent_schedule.operating_days = hbe_vent_sched_prop_ph.operating_days_wk
    new_phx_vent_schedule.operating_weeks = hbe_vent_sched_prop_ph.operating_weeks_year

    for op_period in hbe_vent_sched_prop_ph.daily_operating_periods:
        phx_vent_util_period = getattr(
            new_phx_vent_schedule.operating_periods, op_period.name
        )
        phx_vent_util_period.period_operating_hours = op_period.operation_hours
        phx_vent_util_period.period_operation_speed = op_period.operation_fraction
        setattr(
            new_phx_vent_schedule.operating_periods, op_period.name, phx_vent_util_period
        )

    # -- Keep all the IDs in alignment....
    new_phx_vent_schedule.identifier = hbe_room_prop.ventilation.schedule.identifier
    new_phx_vent_schedule.id_num = new_phx_vent_schedule._count
    hbe_vent_sched_prop_ph.id_num = new_phx_vent_schedule.id_num  # Important!

    return new_phx_vent_schedule


def build_ventilation_schedule_from_hb_room(
    _hb_room: room.Room,
) -> Optional[ventilation.PhxScheduleVentilation]:
    """Build a new Ventilation Schedule based on a Honeybee-Room's energy.ventilation values.

    Arguments:
    ----------
        *_hb_room (room.Room): The Honeybee-Room to get the ventilation pattern data from.

    Returns:
    --------
        * (Optional[ventilation.PhxScheduleVentilation]): The new Ventilation Schedule or None
            if no energy.ventilation or energy.ventilation.schedule is found on the room.
    """

    # -- Make sure that the room has vent schedule
    hbe_prop: RoomEnergyProperties = _hb_room.properties.energy  # type: ignore
    if hbe_prop.ventilation is None:
        return None

    if hbe_prop.ventilation.schedule is None:
        return None

    if _room_has_ph_style_ventilation(_hb_room):
        # -- The room's ventilation DOES have detailed user-inputs for operation
        # -- periods, so just use those. This is the case when a Honeybee-PH
        # -- component is used to build the PH_ScheduleRuleset
        new_vent_schedule = _create_vent_schedule_from_ph_style(_hb_room)
    else:
        # -- There IS a ventilation.schedule, BUT there are no detailed
        # -- Passive House style user-input operation periods set on it. This
        # -- is the case when a normal HB-Ventilation ScheduleRuleset is
        # -- converted over to a PH_ScheduleRuleset
        new_vent_schedule = _create_vent_schedule_from_hb_style(_hb_room)

    # -- Ensure that the operating hours add up to exactly 24
    new_vent_schedule.force_max_utilization_hours()

    return new_vent_schedule


def build_occupancy_schedule_from_hb_room(
    _hb_room: room.Room,
) -> Optional[occupancy.PhxScheduleOccupancy]:
    """Build a new PHX Occupancy Schedule based on a Honeybee-Room's energy.people values.

    Arguments:
    ----------
        * _hb_room (room.Room): The Honeybee Room to build the schedules from.

    Returns:
    --------
        * (Optional[occupancy.PhxScheduleOccupancy]): The new PHX Occupancy Schedule or None.
    """

    # -- Make sure that the room has an occupancy schedule
    hbe_prop: RoomEnergyProperties = _hb_room.properties.energy  # type: ignore
    if hbe_prop.people is None:
        return None

    if hbe_prop.people.occupancy_schedule is None:
        return None

    # -- Aliases
    hbe_schedule = hbe_prop.people.occupancy_schedule
    hbe_schedule_prop_ph: phx_ruleset.ScheduleRulesetPhProperties = hbe_schedule.properties.ph  # type: ignore
    daily_period = hbe_schedule_prop_ph.first_operating_period

    # -- Build the new Schedule
    new_phx_occ_schedule = occupancy.PhxScheduleOccupancy()
    new_phx_occ_schedule.identifier = hbe_schedule.identifier
    new_phx_occ_schedule.display_name = hbe_schedule.display_name
    new_phx_occ_schedule.annual_utilization_days = (
        hbe_schedule_prop_ph.operating_days_year
    )
    new_phx_occ_schedule.relative_utilization_factor = (
        hbe_schedule_prop_ph.annual_average_operating_fraction
    )

    if daily_period:
        new_phx_occ_schedule.start_hour = daily_period.start_hour
        new_phx_occ_schedule.end_hour = daily_period.end_hour
    else:
        new_phx_occ_schedule.start_hour = 0
        new_phx_occ_schedule.end_hour = 24

    return new_phx_occ_schedule


def build_lighting_schedule_from_hb_room(
    _hb_room: room.Room,
) -> Optional[lighting.PhxScheduleLighting]:
    """Build a new PHX Lighting Schedule based on a Honeybee-Room's energy.lighting.schedule values.

    Arguments:
    ----------
        * _hb_room (room.Room): The Honeybee Room to build the schedules from.

    Returns:
    --------
        * (Optional[occupancy.PhxScheduleLighting]): The new PHX Lighting Schedule or None.
    """

    # -- Make sure that the room has an occupancy schedule
    hbe_prop: RoomEnergyProperties = _hb_room.properties.energy  # type: ignore
    if hbe_prop.lighting is None:
        return None

    if hbe_prop.lighting.schedule is None:
        return None

    # -- Aliases
    hbe_schedule = hbe_prop.lighting.schedule
    hbe_schedule_prop_ph: phx_ruleset.ScheduleRulesetPhProperties = hbe_schedule.properties.ph  # type: ignore
    daily_period = hbe_schedule_prop_ph.first_operating_period

    # -- Build the new Schedule
    new_phx_occ_schedule = lighting.PhxScheduleLighting()
    new_phx_occ_schedule.identifier = hbe_schedule.identifier
    new_phx_occ_schedule.display_name = hbe_schedule.display_name
    new_phx_occ_schedule.annual_utilization_days = (
        hbe_schedule_prop_ph.operating_days_year
    )
    new_phx_occ_schedule.relative_utilization_factor = (
        hbe_schedule_prop_ph.annual_average_operating_fraction
    )

    if daily_period:
        new_phx_occ_schedule.start_hour = daily_period.start_hour
        new_phx_occ_schedule.end_hour = daily_period.end_hour
    else:
        new_phx_occ_schedule.start_hour = 0
        new_phx_occ_schedule.end_hour = 24

    return new_phx_occ_schedule


def _add_default_vent_schedule_to_Rooms(_hb_model: model.Model) -> model.Model:
    """Add a default ventilation.schedule to the HB Model's Rooms if they have None.

    Some HB Programs do not have a ventilation.schedule. I *think* this means
    constant operation. So add the constant value (1) default ventilation schedule to the room.

    Arguments:
    ----------
        * _hb_model (model.Model): The Honeybee Model to add the new constant value
            ventilation schedules to.

    Returns:
    --------
        * (model.Model): The Honeybee Model with new ventilation schedules added to
            any Rooms which were missing them.
    """
    type_limit = schedule_type_limit_by_identifier("Fractional")
    default_ventilation_schedule = hbe_ruleset.ScheduleRuleset.from_constant_value(
        "default_schedule", 1.0, type_limit
    )

    for hb_room in _hb_model.rooms:
        if hb_room.properties.energy.ventilation.schedule is None:
            hb_room.properties.energy.ventilation.unlock()
            hb_room.properties.energy.ventilation.schedule = default_ventilation_schedule
            hb_room.properties.energy.ventilation.lock()

    return _hb_model


def add_all_HB_Model_ventilation_schedules_to_PHX_Project(
    _project: project.PhxProject, _hb_model: model.Model
) -> None:
    """Add all the Room's Ventilation Schedules to the project's Collection.

    Arguments:
    ----------
        * _project (project.Project): The PHX-Project to add the new
            Ventilation Schedules to.
        * _hb_model (model.Model): Then Honeybee Model to build up the
            new Ventilation Schedules from.

    Returns:
    --------
        * None
    """

    # -- FIRST: Have to clean up the HB-ventilation schedules where they are missing.
    _hb_model = _add_default_vent_schedule_to_Rooms(_hb_model)

    # -- NEXT: Build up the new Ventilation Schedules from the Room's data
    for hb_room in _hb_model.rooms:
        hbe_room_energy_prop: RoomEnergyProperties = hb_room.properties.energy  # type: ignore
        vent_schedule_id = hbe_room_energy_prop.ventilation.schedule.identifier

        if _project.vent_sched_in_project_collection(vent_schedule_id):
            # -- This is just to help speed things up.
            # -- Don't re-make the util-pattern if it is already in collection.
            continue

        new_phx_vent_schedule = build_ventilation_schedule_from_hb_room(hb_room)
        if new_phx_vent_schedule:
            _project.add_vent_sched_to_collection(new_phx_vent_schedule)


def add_all_HB_Model_occupancy_schedules_to_PHX_Project(
    _project: project.PhxProject, _hb_model: model.Model
) -> None:
    """Add all the Room's Occupancy Schedules to the project's Collection.

    Arguments:
    ----------
        * _project (project.Project): The PHX-Project to add the new
            Occupancy Schedules to.
        * _hb_model (model.Model): Then Honeybee Model to build up the
            new Occupancy Schedules from.

    Returns:
    --------
        * None
    """
    for hb_room in _hb_model.rooms:
        hbe_room_energy_prop: RoomEnergyProperties = hb_room.properties.energy  # type: ignore

        # -- Sometimes there is no 'People'
        if hbe_room_energy_prop.people is None:
            continue

        occ_schedule_id = hbe_room_energy_prop.people.occupancy_schedule.identifier
        if _project.occupancy_sched_in_project_collection(occ_schedule_id):
            # -- This is just to help speed things up.
            # -- Don't re-make the util-pattern if it is already in collection.
            continue

        new_phx_occ_schedule = build_occupancy_schedule_from_hb_room(hb_room)
        _project.add_occupancy_sched_to_collection(new_phx_occ_schedule)


def add_all_HB_Model_lighting_schedules_to_PHX_Project(
    _project: project.PhxProject, _hb_model: model.Model
) -> None:
    for hb_room in _hb_model.rooms:
        hbe_room_energy_prop: RoomEnergyProperties = hb_room.properties.energy  # type: ignore

        if hbe_room_energy_prop.lighting is None:
            continue

        lighting_schedule_id = hbe_room_energy_prop.lighting.schedule.identifier
        if _project.lighting_sched_in_project_collection(lighting_schedule_id):
            # -- This is just to help speed things up.
            # -- Don't re-make the util-pattern if it is already in collection.
            continue

        new_phx_lighting_schedule = build_lighting_schedule_from_hb_room(hb_room)
        _project.add_lighting_sched_to_collection(new_phx_lighting_schedule)


def add_all_HB_schedules_to_PHX_Project(
    _project: project.PhxProject, _hb_model: model.Model
) -> None:
    """Add all the schedules (Ventilation, Occupancy, Lighting) to the PHX Project's Collections.

    Arguments:
    ----------
        * _project (project.Project): The PHX-Project to add the new
            Schedules to.
        * _hb_model (model.Model): Then Honeybee Model to build up the
            new Schedules from.

    Returns:
    --------
        * (None)
    """
    add_all_HB_Model_ventilation_schedules_to_PHX_Project(_project, _hb_model)
    add_all_HB_Model_occupancy_schedules_to_PHX_Project(_project, _hb_model)
    add_all_HB_Model_lighting_schedules_to_PHX_Project(_project, _hb_model)



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Model class for a single PHPP Addition Vent / Space(room)-Entry row."""

from dataclasses import dataclass
from typing import List, Tuple
from functools import partial

from PHX.model import spaces
from PHX.model.schedules import ventilation as phx_vent_sched
from PHX.xl import xl_data
from PHX.xl.xl_data import xl_writable
from PHX.PHPP.phpp_localization import shape_model


@dataclass
class VentSpaceRow:
    """Model class for a single Ventilation Space/Room entry row."""

    __slots__ = ("shape", "phx_room_vent", "phpp_row_ventilator", "phx_vent_pattern")
    shape: shape_model.AddnlVent
    phx_room_vent: spaces.PhxSpace
    phpp_row_ventilator: xl_writable
    phx_vent_pattern: phx_vent_sched.PhxScheduleVentilation

    def _create_range(self, _field_name: str, _row_num: int) -> str:
        """Return the XL Range ("P12",...) for the specific field name."""
        col = getattr(self.shape.rooms.inputs, _field_name).column
        return f"{col}{_row_num}"

    def _get_target_unit(self, _field_name: str) -> str:
        "Return the right target unit for the PHPP item writing (IP | SI)"
        return getattr(self.shape.rooms.inputs, _field_name).unit

    def _get_time_periods(self) -> Tuple[float, float, float]:
        """Return a tuple of the high, standard and min operating periods as % values."""
        periods = self.phx_vent_pattern.operating_periods
        time_high = periods.high.period_operating_hours / 24.0
        time_standard = periods.standard.period_operating_hours / 24.0

        # Ensure total always equals 100%
        time_minimum = 1.0 - (time_high + time_standard)

        return time_high, time_standard, time_minimum

    def _calc_minimum_speed(self) -> float:
        # Weighted average of Min and Basic since PHPP only has 3 values
        periods = self.phx_vent_pattern.operating_periods
        weighted_basic = (
            periods.basic.period_operation_speed * periods.basic.period_operating_hours
        )
        weighed_minimum = (
            periods.minimum.period_operation_speed
            * periods.minimum.period_operating_hours
        )
        total_hours = (
            periods.basic.period_operating_hours + periods.minimum.period_operating_hours
        )

        try:
            return (weighted_basic + weighed_minimum) / total_hours
        except ZeroDivisionError:
            return 0.0

    def _get_speeds(self) -> Tuple[float, float, float]:
        """Return a tuple of the high, standard and minimum speeds as % of max"""
        speed_high = self.phx_vent_pattern.operating_periods.high.period_operation_speed
        speed_standard = (
            self.phx_vent_pattern.operating_periods.standard.period_operation_speed
        )
        speed_minimum = self._calc_minimum_speed()

        return speed_high, speed_standard, speed_minimum

    def create_xl_items(self, _sheet_name: str, _row_num: int) -> List[xl_data.XlItem]:
        """Returns a list of the XL Items to write for this Surface Entry

        Arguments:
        ----------
            * _sheet_name: (str) The name of the worksheet to write to.
            * _row_num: (int) The row number to build the XlItems for
        Returns:
        --------
            * (List[XlItem]): The XlItems to write to the sheet.
        """
        # -- Handle the percentages so always adds up to 100% hours
        time_high, time_standard, time_minimum = self._get_time_periods()
        speed_high, speed_standard, speed_minimum = self._get_speeds()

        # --
        create_range = partial(self._create_range, _row_num=_row_num)
        XLItemAddnlVent = partial(xl_data.XlItem, _sheet_name)
        items: List[xl_data.XlItem] = [
            XLItemAddnlVent(create_range("quantity"), self.phx_room_vent.quantity),
            XLItemAddnlVent(
                create_range("display_name"), f"'{self.phx_room_vent.display_name}"
            ),
            XLItemAddnlVent(create_range("vent_unit_assigned"), self.phpp_row_ventilator),
            XLItemAddnlVent(
                create_range("weighted_floor_area"),
                self.phx_room_vent.floor_area,
                "M2",
                self._get_target_unit("weighted_floor_area"),
            ),
            XLItemAddnlVent(
                create_range("clear_height"),
                self.phx_room_vent.clear_height,
                "M",
                self._get_target_unit("clear_height"),
            ),
            XLItemAddnlVent(
                create_range("V_sup"),
                self.phx_room_vent.ventilation.load.flow_supply,
                "M3/HR",
                self._get_target_unit("V_sup"),
            ),
            XLItemAddnlVent(
                create_range("V_eta"),
                self.phx_room_vent.ventilation.load.flow_extract,
                "M3/HR",
                self._get_target_unit("V_eta"),
            ),
            XLItemAddnlVent(
                create_range("V_trans"),
                self.phx_room_vent.ventilation.load.flow_transfer,
                "M3/HR",
                self._get_target_unit("V_trans"),
            ),
            # -- Operating Days / weeks
            XLItemAddnlVent(
                create_range("operating_hours"), self.phx_vent_pattern.operating_hours
            ),
            XLItemAddnlVent(
                create_range("operating_days"), self.phx_vent_pattern.operating_days
            ),
            XLItemAddnlVent(
                create_range("holiday_days"), self.phx_vent_pattern.holiday_days
            ),
            # -- Operating hours / speeds
            XLItemAddnlVent(create_range("period_high_speed"), speed_high),
            XLItemAddnlVent(create_range("period_high_time"), time_high),
            XLItemAddnlVent(create_range("period_standard_speed"), speed_standard),
            XLItemAddnlVent(create_range("period_standard_time"), time_standard),
            XLItemAddnlVent(create_range("period_minimum_speed"), speed_minimum),
            XLItemAddnlVent(create_range("period_minimum_time"), time_minimum),
        ]

        return items



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""PHX Fresh-Air Ventilation Utilization Schedule."""

from __future__ import annotations
from typing import ClassVar, Any, Union
from dataclasses import dataclass, field
import uuid


@dataclass
class Vent_OperatingPeriod:
    period_operating_hours: float = 0.0  # hours/period
    period_operation_speed: float = 0.0  # % of peak design airflow

    def __eq__(self, other: Vent_OperatingPeriod) -> bool:
        if abs(self.period_operating_hours - other.period_operating_hours) > 0.01:
            return False
        if abs(self.period_operation_speed - other.period_operation_speed) > 0.01:
            return False
        else:
            return True


@dataclass
class Vent_UtilPeriods:
    high: Vent_OperatingPeriod = field(default_factory=Vent_OperatingPeriod)
    standard: Vent_OperatingPeriod = field(default_factory=Vent_OperatingPeriod)
    basic: Vent_OperatingPeriod = field(default_factory=Vent_OperatingPeriod)
    minimum: Vent_OperatingPeriod = field(default_factory=Vent_OperatingPeriod)

    def __eq__(self, other: Vent_UtilPeriods) -> bool:
        if self.high != other.high:
            return False
        if self.standard != other.standard:
            return False
        if self.basic != other.basic:
            return False
        if self.minimum != other.minimum:
            return False
        else:
            return True


@dataclass
class PhxScheduleVentilation:
    """A PHX Schedule for the Ventilation."""

    _count: ClassVar[int] = 0

    id_num: int = field(init=False, default=0)
    name: str = "__unnamed_vent_schedule__"
    identifier: Union[uuid.UUID, str] = field(default_factory=uuid.uuid4)
    operating_hours: float = 24.0
    operating_days: float = 7.0
    operating_weeks: float = 52.0
    operating_periods: Vent_UtilPeriods = field(default_factory=Vent_UtilPeriods)
    holiday_days: float = 0.0

    def __post_init__(self) -> None:
        self.__class__._count += 1
        self.id_num = self.__class__._count

    def force_max_utilization_hours(
        self, _max_hours: float = 24.0, _tol: int = 2
    ) -> None:
        """Ensure that the total utilization hours never exceed target (default=24).
        Will adjust the minimum daily_op_sched as needed.
        """

        b = round(self.operating_periods.standard.period_operating_hours, _tol)
        c = round(self.operating_periods.basic.period_operating_hours, _tol)
        a = round(self.operating_periods.minimum.period_operating_hours, _tol)
        total = a + b + c
        remainder = round(_max_hours - total, _tol)
        self.operating_periods.high.period_operating_hours = remainder

    def __hash__(self):
        return hash(self.identifier)



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Functions to allow the merging of HB-Faces which are used to simplify HB-Models."""

from copy import copy
from collections import defaultdict
import math
from typing import List, Sequence, Tuple, Union, TypeVar

from honeybee.face import Face
from honeybee.shade import Shade
from honeybee.aperture import Aperture
from ladybug_geometry.geometry2d.pointvector import Vector2D
from ladybug_geometry.geometry2d.polygon import Polygon2D
from ladybug_geometry.geometry3d.face import Face3D
from ladybug_geometry.geometry3d.plane import Plane
from ladybug_geometry.geometry3d.pointvector import Point3D, Vector3D


# -----------------------------------------------------------------------------
# -- Geometry Functions
def cross_product(
    a: Union[Sequence[float], Vector3D], b: Union[Sequence[float], Vector3D]
) -> Sequence[float]:
    return [
        a[1] * b[2] - a[2] * b[1],
        a[2] * b[0] - a[0] * b[2],
        a[0] * b[1] - a[1] * b[0],
    ]


def dot_product(
    a: Union[Sequence[float], Vector3D], b: Union[Sequence[float], Vector3D]
) -> float:
    return sum([a[i] * b[i] for i in range(len(a))])


def magnitude(a: Union[Sequence[float], Vector3D]) -> float:
    return sum([a[i] ** 2 for i in range(len(a))]) ** 0.5


def normalize(a: Union[Sequence[float], Vector3D]) -> List[float]:
    mag = magnitude(a)
    return [a[i] / mag for i in range(len(a))]


def angle_between_planes(plane1, plane2, _tolerance):
    # type: (Plane, Plane, float) -> float
    # Calculate the x-axes of the planes
    plane1_x_axis = cross_product(plane1.n, plane1.x)
    plane2_x_axis = cross_product(plane2.n, plane2.x)

    # Normalize the x-axes
    plane1_x_axis = normalize(plane1_x_axis)
    plane2_x_axis = normalize(plane2_x_axis)

    # Calculate the dot product between the x-axes
    dot_product_value = dot_product(plane1_x_axis, plane2_x_axis)

    # Calculate the angle between the x-axes
    # Handle parallel or coincident planes
    if (1.0 - dot_product_value) < _tolerance:
        return 0.0

    angle_rad = math.acos(dot_product_value)

    # Determine the counterclockwise angle
    orientation = cross_product(plane1_x_axis, plane2_x_axis)
    if dot_product(plane1.n, orientation) < 0:
        angle_rad = 2 * math.pi - angle_rad

    return angle_rad


# -----------------------------------------------------------------------------
# Sorting

TFaceOrShade = TypeVar("TFaceOrShade", Face, Shade)


def _hb_face_type_unique_key(_hb_face: TFaceOrShade) -> str:
    """Return a unique key for the HB-Face's type."""

    face_type = str(getattr(_hb_face, "type", "shade"))
    face_bc = str(getattr(_hb_face, "boundary_condition", "shade"))
    hb_prop_e = _hb_face.properties.energy  # type: ignore
    const_name = hb_prop_e.construction.display_name

    return "{}_{}_{}".format(face_type, face_bc, const_name)


def sort_hb_faces_by_type(_faces: List[TFaceOrShade]) -> List[List[TFaceOrShade]]:
    """Group HB-Faces by their type."""

    d = defaultdict(list)
    for face in _faces:
        key = _hb_face_type_unique_key(face)
        d[key].append(face.duplicate())
    return list(d.values())


def sort_faces_by_co_planar(
    _faces: List[TFaceOrShade], _tolerance: float, _angle_tolerance_radians: float
) -> List[List[TFaceOrShade]]:
    """Group HB-Faces with their co-planar neighbors.
    Args:
        _faces: (List[Face | Shade]) A list of HB-Faces to sort.
        _tolerance: (Model units) The tolerance value for co-planarity test, in model units.
        _angle_tolerance: (Radians) The tolerance for co-planarity, in radians.
    Returns:
        (List[List[Face | Shade]]) A list of lists of HB-Faces that are co-planar.
    """

    groups = {}
    for face in _faces:
        for group_plane, group_faces in groups.items():
            if face.geometry.plane.is_coplanar_tolerance(
                group_plane, _tolerance, _angle_tolerance_radians
            ):
                group_faces.append(face)
                break
        else:
            groups[face.geometry.plane] = [face]
    return list(groups.values())


def are_coincident_points(_pt1: Point3D, _pt2: Point3D, _tolerance: float) -> bool:
    """Return True if two Point3D objects are coincident within the tolerance."""
    return _pt1.distance_to_point(_pt2) < _tolerance


def are_touching(_face_2: TFaceOrShade, _face_1: TFaceOrShade, _tolerance: float):
    """Return True if the faces are 'touching' one another within the tolerance."""

    for v in _face_1.vertices:
        if _face_2.geometry.is_point_on_face(v, _tolerance):
            return True
        elif any([are_coincident_points(v, v2, _tolerance) for v2 in _face_2.vertices]):
            return True
    return False


def find_connected_components(
    _hb_faces: List[TFaceOrShade], _tolerance: float
) -> List[List[TFaceOrShade]]:
    """Finds connected components of touching faces in a list of Honeybee faces.

    Args:
        _hb_faces List[Face | Shade]: A list of Honeybee face or shades to search for connected components.
        _tolerance: A tolerance value for determining whether two faces are touching.

    Returns:
        A list of lists, where each inner list contains a connected component of touching faces.
    """

    """Initialize an empty set called visited to keep track of which faces have 
    been visited during the search, and an empty list called components to store 
    the connected component groups."""
    visited = set()
    components = []

    def depth_first_search(node, component):
        # type: (TFaceOrShade, List[TFaceOrShade]) -> None
        """Define a recursive function that takes a starting face node
        and a list component to store the connected component.
        The function adds the starting face to the visited set and the component list,
        and then recursively calls itself on all neighboring faces that are not
        in the visited set and are touching the starting face within
        a given tolerance _tolerance.
        """
        visited.add(node)
        component.append(node)

        for _neighbor_face in _hb_faces:
            if _neighbor_face not in visited and are_touching(
                node, _neighbor_face, _tolerance
            ):
                depth_first_search(_neighbor_face, component)

    """Loop over all the faces in the input list _hb_faces. If a face has not 
    been visited yet, create an empty list called 'component', call the 'depth_first_search'
    function with the face and the new component list, and append the 'component' list 
    to the master components list."""
    for hb_face in _hb_faces:
        if hb_face not in visited:
            component = []
            depth_first_search(hb_face, component)
            components.append(component)

    return components


def sort_hb_faces(
    _hb_faces: List[TFaceOrShade], _tolerance: float, _angle_tolerance_degrees: float
) -> List[List[TFaceOrShade]]:
    """Sort HB-Faces into groups of similar, planar, connected faces.

    Args:
        _hb_faces: (List[Face | Shade]) A list of HB-Faces to sort.
        _tolerance: (Model units) The tolerance value for co-planarity test, in model units.
        _angle_tolerance_degrees: (Degrees) The tolerance for co-planarity, in degrees.
    Returns:
        (List[List[Face | Shade]]) A list of lists of HB-Faces that are similar, planar, and connected.
    """

    face_groups_by_type = sort_hb_faces_by_type(_hb_faces)
    angle_tolerance_radians = math.radians(_angle_tolerance_degrees)

    face_groups_coplanar = []
    for face_group in face_groups_by_type:
        face_groups_coplanar.extend(
            sort_faces_by_co_planar(face_group, _tolerance, angle_tolerance_radians)
        )

    face_groups_connected = []
    for face_group in face_groups_coplanar:
        face_groups_connected.extend(find_connected_components(face_group, _tolerance))

    return face_groups_connected


# -----------------------------------------------------------------------------
# -- Merging Faces


def _get_polygon2d_in_reference_space(
    _polygon2d: Polygon2D, _poly2d_plane: Plane, _base_plane: Plane, _tolerance: float
) -> Polygon2D:
    """Return a Polygon2D in the reference space of the base polygon."""

    # -- Create a Vector2D from each face's origin to the base-geom's origin
    face_origin_in_base_plane_space = _base_plane.xyz_to_xy(_poly2d_plane.o)
    base_plane_origin = copy(_base_plane.xyz_to_xy(_base_plane.o))
    # ------------------------------------------------------------------------
    # -- Construct a Move vector from the face's origin to the base-plane's origin
    mv_x = face_origin_in_base_plane_space.x - base_plane_origin.x
    mv_y = face_origin_in_base_plane_space.y - base_plane_origin.y
    move_vec = Vector2D(mv_x, mv_y)  # type: ignore

    # ------------------------------------------------------------------------
    # -- Move the face's Polygon2D into the base polygon's space
    moved_polygon = _polygon2d.move(move_vec)

    # ------------------------------------------------------------------------
    # -- Rotate the Polygon to align with the base-plane
    angle = angle_between_planes(_base_plane, _poly2d_plane, _tolerance)
    rotated_polygon = moved_polygon.rotate(
        angle=angle, origin=face_origin_in_base_plane_space
    )
    return rotated_polygon


def _add_sub_face(_face: Face, _aperture: Aperture):
    """Add an HB-sub-face (either HB-Aperture or HB-Door) to a parent Face.

    NOTE: this method is copied from honeybee's Grasshopper component "HB Add Subface"
    """
    if isinstance(_aperture, Aperture):  # the sub-face is an Aperture
        _face.add_aperture(_aperture)
    else:  # the sub-face is a Door
        _face.add_door(_aperture)

    return _face


def _check_and_add_sub_face(
    _face: Face,
    _apertures: List[Aperture],
    _tolerance: float,
    _angle_tolerance: float,
):
    """Check whether a HB-sub-face is valid for an HB-face and, if so, add it.

    NOTE: this method is copied from honeybee's Grasshopper component "HB Add Subface"
    """
    for aperture in _apertures:
        if _face.geometry.is_sub_face(aperture.geometry, _tolerance, _angle_tolerance):
            _add_sub_face(_face, aperture)


def _create_new_HB_Face(_face3D: Face3D, _ref_face: Face) -> Face:
    """Create a new HB-Face using a Face3D and a reference HB-Face."""
    new_face = Face(
        identifier=_ref_face.identifier,
        geometry=_face3D,
        type=_ref_face.type,
        boundary_condition=_ref_face.boundary_condition,
    )
    new_face.display_name = _ref_face.display_name
    new_face._user_data = (
        None if _ref_face.user_data is None else _ref_face.user_data.copy()
    )
    new_face._properties._duplicate_extension_attr(_ref_face._properties)

    return new_face


def _create_new_HB_Shade(_face3D: Face3D, _ref_face: Shade) -> Shade:
    """Create a new HB-Shade using a Face3D and a reference HB-Shade."""
    new_face = Shade(
        identifier=_ref_face.identifier,
        geometry=_face3D,
        is_detached=True,
    )
    new_face.display_name = _ref_face.display_name
    new_face._user_data = (
        None if _ref_face.user_data is None else _ref_face.user_data.copy()
    )
    new_face._properties._duplicate_extension_attr(_ref_face._properties)

    return new_face


def _create_new_Face3D(
    _poly2D: Polygon2D, _base_plane: Plane, _ref_face: TFaceOrShade
) -> Face3D:
    """Return a new Face3D based on a Polygon2D and a reference HB-Face."""
    return Face3D(
        boundary=tuple(_base_plane.xy_to_xyz(v) for v in _poly2D.vertices),
        plane=_ref_face.geometry.plane,
    )


def find_parent_and_child_polygons(
    _polygons: List[Polygon2D],
) -> Tuple[List[Polygon2D], List[Polygon2D]]:
    """Return a Tuple of any parent (container) and child (contained) polygons from a list of polygons."""

    # Initialize empty lists for parent and child surfaces
    parent_polygon = []
    child_polygons = []

    # Compare each surface with all other surfaces
    for i, _polygon in enumerate(_polygons):
        is_inside_any = False

        for j, other_polygon in enumerate(_polygons):
            if i == j:
                continue

            if _polygon.is_polygon_inside(other_polygon):
                is_inside_any = True
                break

        # If the surface is inside any other surface, add it to child_surfaces
        # Otherwise, it is a parent surface
        if is_inside_any:
            parent_polygon.append(_polygon)
        else:
            child_polygons.append(_polygon)

    return parent_polygon, child_polygons


def merge_hb_face_polygons(
    _faces: List[TFaceOrShade], _tolerance: float, _angle_tolerance_degrees: float
) -> Tuple[List[Polygon2D], Plane, TFaceOrShade]:
    # -------------------------------------------------------------------------
    # -- This will be the reference face for everything else to match
    reference_face: TFaceOrShade = _faces.pop(0).duplicate()  # type: ignore
    reference_plane: Plane = copy(reference_face.geometry.plane)
    polygons_in_ref_space: List[Polygon2D] = []
    polygons_in_ref_space.append(reference_face.geometry.polygon2d)

    # -------------------------------------------------------------------------
    # -- Get all the Polygon2Ds in the same reference space
    poly2ds = [f.geometry.polygon2d for f in _faces]
    planes = (f.geometry.plane for f in _faces)

    if len(poly2ds) > 100:
        print(
            f"Merging together {len(poly2ds)} polygons for '{reference_face.display_name}'."
        )
        print("Consider reducing the complexity of the geometry.")

    for poly2d_plane, poly2d in zip(planes, poly2ds):
        polygons_in_ref_space.append(
            _get_polygon2d_in_reference_space(
                poly2d, poly2d_plane, reference_plane, _tolerance
            )
        )

    # -------------------------------------------------------------------------
    # -- Try and merge all the new Polygon2Ds together.
    merged_polygons = Polygon2D.boolean_union_all(polygons_in_ref_space, _tolerance)

    if len(poly2ds) > 100:
        print(f"merge_hb_face_polygons resulted in: {len(merged_polygons)} faces.")

    return (merged_polygons, reference_plane, reference_face)


def merge_hb_faces(
    _faces: List[Face], _tolerance: float, _angle_tolerance_degrees: float
):
    """Merge a group of HB-Faces into the fewest number of faces possible."""

    if not _faces:
        return []

    if len(_faces) == 1:
        return _faces

    # -------------------------------------------------------------------------
    # -- Before anything else, preserve any the Apertures for adding back in later
    apertures = []
    for f in _faces:
        apertures.extend([ap.duplicate() for ap in getattr(f, "apertures", [])])

    # -------------------------------------------------------------------------
    # -- Merge the Polygons togther
    merged_polygons, ref_plane, ref_face = merge_hb_face_polygons(
        _faces, _tolerance, _angle_tolerance_degrees
    )

    # -------------------------------------------------------------------------
    # -- Create new Face3D, and HB-Faces from the Polygon2Ds
    faces = []
    if len(merged_polygons) == 1:
        # -- Create new faces for the merged Polygon2Ds
        face3ds = (_create_new_Face3D(p, ref_plane, ref_face) for p in merged_polygons)
        faces = [_create_new_HB_Face(f3d, ref_face) for f3d in face3ds]
    elif len(merged_polygons) > 1:
        # -- It may mean that there are 'holes' in a surface? So try and find
        # -- the parent and any child surfaces.

        parent_polygon, child_polygons = find_parent_and_child_polygons(merged_polygons)

        # -- Check the results
        if len(parent_polygon) != 1:
            # -- Something went wrong, give up.
            _faces.append(ref_face)
            return _faces

        # -- If only 1 parent, lets make some Face3Ds and Faces
        parent_face_3d = [
            _create_new_Face3D(p, ref_plane, ref_face) for p in parent_polygon
        ]
        child_face_3ds = [
            _create_new_Face3D(p, ref_plane, ref_face) for p in child_polygons
        ]
        face_3ds = [Face3D.from_punched_geometry(parent_face_3d[0], child_face_3ds)]
        faces = [_create_new_HB_Face(f3d, ref_face) for f3d in face_3ds]

    # -------------------------------------------------------------------------
    # -- Add the apertures back in
    faces_with_apertures_ = []
    for _face in faces:
        _check_and_add_sub_face(_face, apertures, _tolerance, _angle_tolerance_degrees)
        faces_with_apertures_.append(_face)

    return faces_with_apertures_


def merge_hb_shades(
    _faces: List[Shade], _tolerance: float, _angle_tolerance_degrees: float
) -> List[Shade]:
    """Merge a group of HB-Shades into the fewest number of shades possible."""
    if not _faces:
        return []

    if len(_faces) == 1:
        return _faces

    # -------------------------------------------------------------------------
    # -- Merge the Polygons togther
    merged_polygons, ref_plane, ref_face = merge_hb_face_polygons(
        _faces, _tolerance, _angle_tolerance_degrees
    )

    # -------------------------------------------------------------------------
    # -- Create new Face3D, and HB-Faces from the Polygon2Ds
    hb_shades_ = []
    if len(merged_polygons) == 1:
        # -- Create new faces for the merged Polygon2Ds
        face3ds = (_create_new_Face3D(p, ref_plane, ref_face) for p in merged_polygons)
        hb_shades_ = [_create_new_HB_Shade(f3d, ref_face) for f3d in face3ds]
    elif len(merged_polygons) > 1:
        # -- It may mean that there are 'holes' in a surface? So try and find
        # -- the parent and any child surfaces.

        parent_polygon, child_polygons = find_parent_and_child_polygons(merged_polygons)

        # -- Check the results
        if len(parent_polygon) != 1:
            # -- Something went wrong, give up.
            _faces.append(ref_face)
            return _faces

        # -- If only 1 parent, lets make some Face3Ds and Faces
        parent_face_3d = [
            _create_new_Face3D(p, ref_plane, ref_face) for p in parent_polygon
        ]
        child_face_3ds = [
            _create_new_Face3D(p, ref_plane, ref_face) for p in child_polygons
        ]
        face_3ds = Face3D.from_punched_geometry(parent_face_3d[0], child_face_3ds)
        hb_shades_ = [_create_new_HB_Shade(f3d, ref_face) for f3d in face_3ds]

    return hb_shades_



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Controller Class for the PHPP "Electricity non-res" worksheet."""

from __future__ import annotations
from typing import List, Optional, Generator, Tuple

from PHX.xl import xl_app, xl_data
from PHX.PHPP.phpp_localization import shape_model
from PHX.PHPP.phpp_model.elec_non_res import ExistingLightingRow


class Lighting:
    def __init__(self, _xl: xl_app.XLConnection, _shape: shape_model.ElecNonRes) -> None:
        self.xl = _xl
        self.shape = _shape
        self._section_header_row: Optional[int] = None
        self._section_first_entry_row: Optional[int] = None
        self._section_last_entry_row: Optional[int] = None

    @property
    def section_header_row(self) -> int:
        """Return the row number of the 'Lighting' section header."""
        if not self._section_header_row:
            self._section_header_row = self.find_section_header_row()
        return self._section_header_row

    @property
    def section_first_entry_row(self) -> int:
        """Return the row number of the very first user-input entry row in the 'Lighting' section."""
        if not self._section_first_entry_row:
            self._section_first_entry_row = self.find_section_first_entry_row()
        return self._section_first_entry_row

    @property
    def section_last_entry_row(self) -> int:
        """Return the row number of the last user-input entry row in the 'Lighting' section."""
        if not self._section_last_entry_row:
            self._section_last_entry_row = self.find_section_last_entry_row()
        return self._section_last_entry_row

    def find_section_header_row(self, _row_start: int = 1, _row_end: int = 100) -> int:
        """Return the row number of the 'Lighting input' section header."""

        xl_data = self.xl.get_single_column_data(
            _sheet_name=self.shape.name,
            _col=self.shape.lighting_rows.locator_col_header,
            _row_start=_row_start,
            _row_end=_row_end,
        )

        for i, val in enumerate(xl_data, start=1):
            if val == self.shape.lighting_rows.locator_string_header:
                return i

        raise Exception(
            f'\n\tError: Not able to find the "Lighting input" input section of '
            f'the "{self.shape.name}" worksheet? Please be sure the section begins '
            f'with the "{self.shape.lighting_rows.locator_string_header}" flag in '
            f"column {self.shape.lighting_rows.locator_col_header}."
        )

    def find_section_first_entry_row(self) -> int:
        """Return the row number of the very first user-input entry row in the 'Lighting input' section."""

        xl_data = self.xl.get_single_column_data(
            _sheet_name=self.shape.name,
            _col=self.shape.lighting_rows.locator_col_entry,
            _row_start=self.section_header_row,
            _row_end=self.section_header_row + 25,
        )

        for i, val in enumerate(xl_data, start=self.section_header_row):
            if val == self.shape.lighting_rows.locator_string_entry:
                return i + 1

        raise Exception(
            f'\n\tError: Not able to find the first surface entry row in the "Lighting input" section?'
        )

    def find_section_last_entry_row(self, _start_row: Optional[int] = None) -> int:
        """Return the row number of the last user-input entry row in the 'Lighting input' section."""

        if not _start_row:
            _start_row = self.section_first_entry_row
        elif _start_row > 10_000:
            raise Exception(
                f'\n\tError: Not able to find the last surface entry row in the "Lighting input" section?'
            )

        _row_end = _start_row + 500
        xl_data = self.xl.get_single_column_data(
            _sheet_name=self.shape.name,
            _col=self.shape.lighting_rows.locator_col_entry,
            _row_start=_start_row,
            _row_end=_row_end,
        )

        for i, val in enumerate(xl_data, start=self.section_first_entry_row):
            if val == self.shape.lighting_rows.locator_string_exit:
                return i - 2
        else:
            return self.find_section_last_entry_row(_row_end)

    def get_lighting_row_data(self, _row_num: int) -> ExistingLightingRow:
        """Return the lighting row object for the given row number."""
        return ExistingLightingRow(
            self.shape,
            self.xl.get_single_row_data(self.shape.name, _row_num),
        )

    @property
    def all_lighting_row_data(
        self,
    ) -> Generator[Tuple[int, ExistingLightingRow], None, None]:
        """Return a generator of all the row_nums and lighting rows in the Lighting worksheet.

        Yields:
        -------
            * Tuple[int, ExistingLightingRow]: The row-number and the surface row object.
        """

        for i in range(self.section_first_entry_row, self.section_last_entry_row):
            exg_surface_row = self.get_lighting_row_data(i)
            yield i, exg_surface_row

    @property
    def used_lighting_row_numbers(self) -> Generator[int, None, None]:
        """Return a generator of all the row_nums that have data in the Lighting section."""
        data = self.xl.get_single_column_data(
            self.shape.name,
            str(self.shape.lighting_rows.inputs.room_zone_name),
            self.section_first_entry_row,
            self.section_last_entry_row,
        )
        return (
            i
            for i, val in enumerate(data, start=self.section_first_entry_row)
            if val is not None
        )

    def _get_target_unit(self, _field_name: str) -> str:
        "Return the right target unit for the PHPP item writing (IP | SI)"
        return getattr(self.shape.lighting_rows.inputs, _field_name).unit

    def set_lighting_power_density(
        self, _row_num: int, _power_density: float, _unit: str = "W/M2"
    ) -> None:
        """Set the lighting power density for the given row number."""
        _range = f"{self.shape.lighting_rows.inputs.installed_power.column}{_row_num}"
        self.xl.write_xl_item(
            xl_data.XlItem(
                sheet_name=self.shape.name,
                xl_range=_range,
                write_value=_power_density,
                input_unit=_unit,
                target_unit=self._get_target_unit("installed_power"),
            )
        )


class ElecNonRes:
    """IO Controller for the PHPP "Electricity non-res" worksheet."""

    def __init__(self, _xl: xl_app.XLConnection, _shape: shape_model.ElecNonRes) -> None:
        self.xl = _xl
        self.shape = _shape
        self.lighting = Lighting(self.xl, self.shape)



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Controller Classes for the PHPP 'Cooling Load' (Peak Cooling Load) Worksheet."""

from __future__ import annotations
from typing import Dict, Any

from ph_units.unit_type import Unit

from PHX.xl import xl_app
from PHX.PHPP.phpp_localization import shape_model


class CoolingPeakLoad:
    """IO Controller Classes for the PHPP 'Cooling Load' (Peak Cooling Load) Worksheet."""

    def __init__(self, _xl: xl_app.XLConnection, _shape: shape_model.CoolingPeakLoad):
        self.xl = _xl
        self.shape = _shape

    def _get_peak_load(self, _col: str) -> Dict[str, Unit]:
        """Return a Dict of all the Peak Heating Load data from the specified column."""
        shp = self.shape

        pT = self.xl.get_single_data_item(
            shp.name, f"{_col}{shp.row_total_losses_transmission}"
        )
        pV = self.xl.get_single_data_item(
            shp.name, f"{_col}{shp.row_total_losses_ventilation}"
        )
        pS = self.xl.get_single_data_item(shp.name, f"{_col}{shp.row_total_gains_solar}")
        pI = self.xl.get_single_data_item(
            shp.name, f"{_col}{shp.row_total_gains_internal}"
        )
        pC = self.xl.get_single_data_item(
            shp.name, f"{_col}{self.shape.row_total_sensible_load}"
        )
        pD = self.xl.get_single_data_item(
            shp.name, f"{_col}{self.shape.row_total_latent_load}"
        )
        return {
            "losses_transmission": Unit(float(pT or 0.0), str(shp.unit)),
            "losses_ventilation": Unit(float(pV or 0.0), str(shp.unit)),
            "gains_solar": Unit(float(pS or 0.0), str(shp.unit)),
            "gains_internal": Unit(float(pI or 0.0), str(shp.unit)),
            "peak_sensible_cooling_load": Unit(float(pC or 0.0), str(shp.unit)),
            "peak_latent_cooling_load": Unit(float(pD or 0.0), str(shp.unit)),
        }

    def get_peak_load_1_data(self) -> Dict[str, Any]:
        """Return a Dict of all the Weather-1 Peak Heating Load data (Watts | Btuh)"""
        return self._get_peak_load(self.shape.col_weather_1)

    def get_peak_load_2_data(self) -> Dict[str, Any]:
        """Return a Dict of all the Weather-2 Peak Heating Load data (Watts | Btuh)"""
        return self._get_peak_load(self.shape.col_weather_2)
